/**
 * The MIT License (MIT)
 * Copyright (c) 2017-present Dmitry Soshnikov <dmitry.soshnikov@gmail.com>
 */

'use strict';

/**
 * Helper `gen` function calls node type handler.
 */
function gen(node) {
  return node ? generator[node.type](node) : '';
}

/**
 * AST handler.
 */
var generator = {
  RegExp: function RegExp(node) {
    return "/".concat(gen(node.body), "/").concat(node.flags);
  },
  Alternative: function Alternative(node) {
    return (node.expressions || []).map(gen).join('');
  },
  Disjunction: function Disjunction(node) {
    return "".concat(gen(node.left), "|").concat(gen(node.right));
  },
  Group: function Group(node) {
    var expression = gen(node.expression);
    if (node.capturing) {
      // A named group.
      if (node.name) {
        return "(?<".concat(node.nameRaw || node.name, ">").concat(expression, ")");
      }
      return "(".concat(expression, ")");
    }
    return "(?:".concat(expression, ")");
  },
  Backreference: function Backreference(node) {
    switch (node.kind) {
      case 'number':
        return "\\".concat(node.reference);
      case 'name':
        return "\\k<".concat(node.referenceRaw || node.reference, ">");
      default:
        throw new TypeError("Unknown Backreference kind: ".concat(node.kind));
    }
  },
  Assertion: function Assertion(node) {
    switch (node.kind) {
      case '^':
      case '$':
      case '\\b':
      case '\\B':
        return node.kind;
      case 'Lookahead':
        {
          var assertion = gen(node.assertion);
          if (node.negative) {
            return "(?!".concat(assertion, ")");
          }
          return "(?=".concat(assertion, ")");
        }
      case 'Lookbehind':
        {
          var _assertion = gen(node.assertion);
          if (node.negative) {
            return "(?<!".concat(_assertion, ")");
          }
          return "(?<=".concat(_assertion, ")");
        }
      default:
        throw new TypeError("Unknown Assertion kind: ".concat(node.kind));
    }
  },
  CharacterClass: function CharacterClass(node) {
    var expressions = node.expressions.map(gen).join('');
    if (node.negative) {
      return "[^".concat(expressions, "]");
    }
    return "[".concat(expressions, "]");
  },
  ClassRange: function ClassRange(node) {
    return "".concat(gen(node.from), "-").concat(gen(node.to));
  },
  Repetition: function Repetition(node) {
    return "".concat(gen(node.expression)).concat(gen(node.quantifier));
  },
  Quantifier: function Quantifier(node) {
    var quantifier;
    var greedy = node.greedy ? '' : '?';
    switch (node.kind) {
      case '+':
      case '?':
      case '*':
        quantifier = node.kind;
        break;
      case 'Range':
        // Exact: {1}
        if (node.from === node.to) {
          quantifier = "{".concat(node.from, "}");
        }
        // Open: {1,}
        else if (!node.to) {
          quantifier = "{".concat(node.from, ",}");
        }
        // Closed: {1,3}
        else {
          quantifier = "{".concat(node.from, ",").concat(node.to, "}");
        }
        break;
      default:
        throw new TypeError("Unknown Quantifier kind: ".concat(node.kind));
    }
    return "".concat(quantifier).concat(greedy);
  },
  Char: function Char(node) {
    var value = node.value;
    switch (node.kind) {
      case 'simple':
        {
          if (node.escaped) {
            return "\\".concat(value);
          }
          return value;
        }
      case 'hex':
      case 'unicode':
      case 'oct':
      case 'decimal':
      case 'control':
      case 'meta':
        return value;
      default:
        throw new TypeError("Unknown Char kind: ".concat(node.kind));
    }
  },
  UnicodeProperty: function UnicodeProperty(node) {
    var escapeChar = node.negative ? 'P' : 'p';
    var namePart;
    if (!node.shorthand && !node.binary) {
      namePart = "".concat(node.name, "=");
    } else {
      namePart = '';
    }
    return "\\".concat(escapeChar, "{").concat(namePart).concat(node.value, "}");
  }
};
module.exports = {
  /**
   * Generates a regexp string from an AST.
   *
   * @param Object ast - an AST node
   */
  generate: gen
};